# Copyright 2025, NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path
from typing import Any, Dict

from genai_perf.config.input.base_config import BaseConfig
from genai_perf.config.input.config_defaults import OutputDefaults
from genai_perf.config.input.config_field import ConfigField


class ConfigOutput(BaseConfig):
    """
    Describes the configuration output options
    """

    def __init__(self) -> None:
        super().__init__()
        self.artifact_directory: Any = ConfigField(
            default=Path(OutputDefaults.ARTIFACT_DIRECTORY),
            verbose_template_comment="The directory to store all the (output) artifacts generated by\
                    \nGenAI-Perf and Perf Analyzer.",
        )
        self.checkpoint_directory: Any = ConfigField(
            default=Path(OutputDefaults.CHECKPOINT_DIRECTORY),
            verbose_template_comment="The directory to store/restore the checkpoint generated by GenAI-Perf.",
        )
        self.profile_export_file: Any = ConfigField(
            default=Path(OutputDefaults.PROFILE_EXPORT_FILE),
            verbose_template_comment="The path where Perf Analyzer profiling data will be exported.\
                \nBy default, the profile export will be to profile_export.json.\
                \nBy default, the GenAI-Perf export will be to <profile_export_file>_genai_perf.csv",
        )
        self.generate_plots: Any = ConfigField(
            default=OutputDefaults.GENERATE_PLOTS,
            verbose_template_comment="Enables the generation of plots",
        )

        self.enable_checkpointing: Any = ConfigField(
            default=OutputDefaults.ENABLE_CHECKPOINTING,
            verbose_template_comment="Enables the checkpointing of the GenAI-Perf state.\
                \nThis is useful for running GenAI-Perf in a stateful manner.",
        )

    def parse(self, output: Dict[str, Any]) -> None:
        for key, value in output.items():
            if key == "artifact_directory":
                self.artifact_directory = Path(value)
            elif key == "checkpoint_directory":
                self.checkpoint_directory = Path(value)
            elif key == "profile_export_file":
                self.profile_export_file = Path(value)
            elif key == "generate_plots":
                self.generate_plots = value
            elif key == "enable_checkpointing":
                self.enable_checkpointing = value
            else:
                raise ValueError(f"User Config: {key} is not a valid output parameter")
